import datetime

from sqlalchemy import Column, Integer, String, DateTime, Boolean, Float, BigInteger, Date
from pagexdata.database import Base
from pydantic import BaseModel
from typing import Optional


class RedditHits(Base):
    __tablename__ = 'reddit_hits'

    import_date = Column(Date, primary_key=True)
    ticker_symbol = Column(String, primary_key=True)
    subreddit = Column(String, primary_key=True)
    hits = Column(Integer)
    positive_hits = Column(Boolean)
    rank = Column(Integer, nullable=True)
    previous_rank = Column(Integer, nullable=True)
    change_rank = Column(Integer, nullable=True)
    percentage_ticker_to_total = Column(Float, nullable=True)
    change_hits_one_day = Column(Integer, nullable=True)
    change_hits_two_days = Column(Integer, nullable=True)
    change_hits_three_days = Column(Integer, nullable=True)
    change_hits_one_week = Column(Integer, nullable=True)
    change_hits_two_weeks = Column(Integer, nullable=True)
    change_hits_four_weeks = Column(Integer, nullable=True)
    hits_volatility_one_week = Column(Float, nullable=True)
    hits_volatility_two_weeks = Column(Float, nullable=True)


class StockData(Base):
    __tablename__ = 'stockdata'

    import_date = Column(Date, primary_key=True)
    ticker_symbol = Column(String, primary_key=True)  # symbol
    price = Column(Float)  # regularMarketPrice
    volume = Column(BigInteger)  # regularMarketVolume
    averageDailyVolume10Day = Column(BigInteger, nullable=True)  # averageDailyVolume10Day
    averageDailyVolume3Month = Column(BigInteger, nullable=True)  # averageDailyVolume3Month
    bookValue = Column(Float, nullable=True)  # https://www.investopedia.com/articles/investing/110613/market-value-versus-book-value.asp
    currency = Column(String, nullable=True)
    displayName = Column(String, nullable=True)
    epsTrailingTwelveMonths = Column(Float, nullable=True)  # https://www.investopedia.com/ask/answers/070114/what-formula-calculating-earnings-share-eps.asp
    epsCurrentYear = Column(Float, nullable=True)  # https://www.investopedia.com/terms/e/eps.asp
    epsForward = Column(Float, nullable=True)  # https://www.investopedia.com/ask/answers/070114/what-formula-calculating-earnings-share-eps.asp
    priceEpsCurrentYear = Column(Float, nullable=True)  # https://www.investopedia.com/terms/p/price-earningsratio.asp
    trailingPE = Column(Float, nullable=True)  # https://www.investopedia.com/terms/t/trailingpe.asp
    forwardPE = Column(Float, nullable=True)  # https://www.investopedia.com/terms/f/forwardpe.asp
    esgPopulated = Column(Boolean, nullable=True)  # https://finance.yahoo.com/news/esg-investing-135435590.html
    fiftyDayAverage = Column(Float, nullable=True)  # SMA?
    fiftyDayAverageChange = Column(Float, nullable=True)  # SMA change?
    fiftyDayAverageChangePercent = Column(Float, nullable=True)  # SMA change percent?
    twoHundredDayAverage = Column(Float, nullable=True)  # SMA?
    twoHundredDayAverageChange = Column(Float, nullable=True)  # SMA change?
    twoHundredDayAverageChangePercent = Column(Float, nullable=True)  # SMA change percent?
    fiftyTwoWeekHigh = Column(Float, nullable=True)
    fiftyTwoWeekHighChange = Column(Float, nullable=True)
    fiftyTwoWeekHighChangePercent = Column(Float, nullable=True)
    fiftyTwoWeekLow = Column(Float, nullable=True)
    fiftyTwoWeekLowChange = Column(Float, nullable=True)
    fiftyTwoWeekLowChangePercent = Column(Float, nullable=True)
    fiftyTwoWeekRange = Column(String, nullable=True)
    longName = Column(String, nullable=True)
    market = Column(String, nullable=True)
    marketCap = Column(BigInteger, nullable=True)  # https://www.investopedia.com/investing/market-capitalization-defined/
    priceToBook = Column(Float, nullable=True)  # https://www.investopedia.com/investing/using-price-to-book-ratio-evaluate-companies/
    quoteType = Column(String, nullable=True)
    region = Column(String, nullable=True)
    regularMarketChange = Column(Float, nullable=True)
    regularMarketChangePercent = Column(Float, nullable=True)
    regularMarketDayHigh = Column(Float, nullable=True)
    regularMarketDayLow = Column(Float, nullable=True)
    regularMarketDayRange = Column(String, nullable=True)
    regularMarketOpen = Column(Float, nullable=True)
    regularMarketPreviousClose = Column(Float, nullable=True)
    sharesOutstanding = Column(BigInteger, nullable=True)  # https://www.investopedia.com/terms/o/outstandingshares.asp
    trailingAnnualDividendRate = Column(Float, nullable=True)  # https://www.investopedia.com/terms/d/dividendyield.asp
    trailingAnnualDividendYield = Column(Float, nullable=True)  # https://www.investopedia.com/terms/d/dividendyield.asp


class ImportLog(Base):
    __tablename__ = 'import_log'

    log_datetime = Column(DateTime, primary_key=True)
    type = Column(String, primary_key=True)
    trigger_type = Column(String)
    state = Column(String)
    import_date = Column(Date, nullable=False)
    message = Column(String, nullable=True)
    extra = Column(String, nullable=True)


class CalculationLog(Base):
    __tablename__ = 'calculation_log'

    log_datetime = Column(DateTime, primary_key=True)
    type = Column(String, primary_key=True)
    trigger_type = Column(String)
    state = Column(String)
    calculation_date = Column(Date, nullable=False)
    message = Column(String, nullable=True)
    extra = Column(String, nullable=True)


class DeletionLog(Base):
    __tablename__ = 'deletion_log'

    log_datetime = Column(DateTime, primary_key=True)
    trigger_type = Column(String)
    state = Column(String)
    message = Column(String, nullable=True)
    extra = Column(String, nullable=True)


# pagination entities for rest api
class RedditHitsPage(BaseModel):
    import_date: datetime.date
    ticker_symbol: str
    subreddit: str
    hits: int
    rank: Optional[int] = None
    previous_rank: Optional[int] = None
    change_rank: Optional[int] = None
    percentage_ticker_to_total: Optional[float] = None
    change_hits_one_day: Optional[int] = None
    change_hits_two_days: Optional[int] = None
    change_hits_three_days: Optional[int] = None
    change_hits_one_week: Optional[int] = None
    change_hits_two_weeks: Optional[int] = None
    change_hits_four_weeks: Optional[int] = None
    hits_volatility_one_week: Optional[float] = None
    hits_volatility_two_weeks: Optional[float] = None

    class Config:
        orm_mode = True


class StockDataPage(BaseModel):
    import_date: datetime.date
    ticker_symbol: str  # symbol
    price: Optional[float]  # regularMarketPrice
    volume: Optional[int]  # regularMarketVolume
    averageDailyVolume10Day: Optional[int]  # averageDailyVolume10Day
    averageDailyVolume3Month: Optional[int]  # averageDailyVolume3Month
    bookValue: Optional[float]  # https://www.investopedia.com/articles/investing/110613/market-value-versus-book-value.asp
    currency: Optional[str]
    displayName: Optional[str]
    epsTrailingTwelveMonths: Optional[float]  # https://www.investopedia.com/ask/answers/070114/what-formula-calculating-earnings-share-eps.asp
    epsCurrentYear: Optional[float]  # https://www.investopedia.com/terms/e/eps.asp
    epsForward: Optional[float]  # https://www.investopedia.com/ask/answers/070114/what-formula-calculating-earnings-share-eps.asp
    priceEpsCurrentYear: Optional[float]  # https://www.investopedia.com/terms/p/price-earningsratio.asp
    trailingPE: Optional[float]  # https://www.investopedia.com/terms/t/trailingpe.asp
    forwardPE: Optional[float]  # https://www.investopedia.com/terms/f/forwardpe.asp
    esgPopulated: Optional[bool]  # https://finance.yahoo.com/news/esg-investing-135435590.html
    fiftyDayAverage: Optional[float]  # SMA?
    fiftyDayAverageChange: Optional[float]  # SMA change?
    fiftyDayAverageChangePercent: Optional[float]  # SMA change percent?
    twoHundredDayAverage: Optional[float]  # SMA?
    twoHundredDayAverageChange: Optional[float]  # SMA change?
    twoHundredDayAverageChangePercent: Optional[float]  # SMA change percent?
    fiftyTwoWeekHigh: Optional[float]
    fiftyTwoWeekHighChange: Optional[float]
    fiftyTwoWeekHighChangePercent: Optional[float]
    fiftyTwoWeekLow: Optional[float]
    fiftyTwoWeekLowChange: Optional[float]
    fiftyTwoWeekLowChangePercent: Optional[float]
    fiftyTwoWeekRange: Optional[str]
    longName: Optional[str]
    market: Optional[str]
    marketCap: Optional[int]  # https://www.investopedia.com/investing/market-capitalization-defined/
    priceToBook: Optional[float]  # https://www.investopedia.com/investing/using-price-to-book-ratio-evaluate-companies/
    quoteType: Optional[str]
    region: Optional[str]
    regularMarketChange: Optional[float]
    regularMarketChangePercent: Optional[float]
    regularMarketDayHigh: Optional[float]
    regularMarketDayLow: Optional[float]
    regularMarketDayRange: Optional[str]
    regularMarketOpen: Optional[float]
    regularMarketPreviousClose: Optional[float]
    sharesOutstanding: Optional[int]  # https://www.investopedia.com/terms/o/outstandingshares.asp
    trailingAnnualDividendRate: Optional[float]  # https://www.investopedia.com/terms/d/dividendyield.asp
    trailingAnnualDividendYield: Optional[float]  # https://www.investopedia.com/terms/d/dividendyield.asp

    class Config:
        orm_mode = True


class ImportLogPage(BaseModel):
    log_datetime: datetime.datetime
    type: str
    trigger_type: str
    state: str
    import_date: datetime.date
    message: Optional[str] = None
    extra: Optional[str] = None

    class Config:
        orm_mode = True


class CalculationLogPage(BaseModel):
    log_datetime: datetime.datetime
    type: str
    trigger_type: str
    state: str
    calculation_date: datetime.date
    message: Optional[str] = None
    extra: Optional[str] = None

    class Config:
        orm_mode = True


class DeletionLogPage(BaseModel):
    log_datetime: datetime.datetime
    trigger_type: str
    state: str
    message: Optional[str] = None
    extra: Optional[str] = None

    class Config:
        orm_mode = True



from sqlalchemy import update
import pandas as pd
from pandas import DataFrame
from .datetime_service import date_format


def create(entity) -> None:
    from pagexdata.database import db_session
    try:
        db_session.add(entity)
        db_session.commit()
        db_session.close()
    except Exception as ex:
        db_session.rollback()
        raise ex


def create_all(entity_list) -> None:
    from pagexdata.database import db_session
    try:
        db_session.add_all(entity_list)
        db_session.commit()
        db_session.close()
    except Exception as ex:
        db_session.rollback()
        raise ex


def read_all(entity_class_):
    from pagexdata.database import db_session
    result = db_session.query(entity_class_).all()
    db_session.close()
    return result


def read_all_by(entity_class_, query_filter=None, query_order_by=None, order_by_desc=False):
    from pagexdata.database import db_session
    query = db_session.query(entity_class_)
    if query_filter is not None:
        query = query.filter(*query_filter)
    if query_order_by is not None and order_by_desc:
        query = query.order_by(query_order_by.desc())
    elif query_order_by is not None and not order_by_desc:
        query = query.order_by(query_order_by)
    result = query.all()
    db_session.close()
    return result


def read_all_as_dataframe(entity_class_) -> DataFrame:
    from pagexdata.database import url
    return pd.read_sql(entity_class_.__tablename__,
                       url.render_as_string(hide_password=False),
                       parse_dates={"import_date": {"format": "%y-%m-%d"}})  # date_format from datetime_service should be used here, but somehow the format does not seem to work here...


def delete_all(entity_class_) -> None:
    from pagexdata.database import db_session
    try:
        result = db_session.query(entity_class_).all()
        for element in result:
            db_session.delete(element)
            db_session.commit()
            db_session.close()
    except Exception as ex:
        db_session.rollback()
        raise ex


def delete_all_by(entity_class_, query_filter=None):
    from pagexdata.database import db_session
    try:
        query = db_session.query(entity_class_)
        if query_filter is not None:
            query = query.filter(*query_filter)
        query.delete(synchronize_session=False)
        db_session.commit()
        db_session.close()
    except Exception as ex:
        db_session.rollback()
        raise ex


def update_reddit_hits_entry(entity) -> None:
    from pagexdata.database import db_session
    try:
        from pagexdata.entities import RedditHits
        db_session.execute(update(RedditHits)
           .where(RedditHits.import_date == entity.import_date)
           .where(RedditHits.subreddit == entity.subreddit)
           .where(RedditHits.ticker_symbol == entity.ticker_symbol)
           .values(rank=entity.rank,
                   positive_hits=entity.positive_hits,
                   previous_rank=entity.previous_rank,
                   change_rank=entity.change_rank,
                   percentage_ticker_to_total=entity.percentage_ticker_to_total,
                   change_hits_one_day=entity.change_hits_one_day,
                   change_hits_two_days=entity.change_hits_two_days,
                   change_hits_three_days=entity.change_hits_three_days,
                   change_hits_one_week=entity.change_hits_one_week,
                   change_hits_two_weeks=entity.change_hits_two_weeks,
                   change_hits_four_weeks=entity.change_hits_four_weeks,
                   hits_volatility_one_week=entity.hits_volatility_one_week,
                   hits_volatility_two_weeks=entity.hits_volatility_two_weeks))
        db_session.commit()
        db_session.close()
    except Exception as ex:
        db_session.rollback()
        raise ex


def update_all_reddit_hits_rankings_by_date(import_date, subreddits, ticker_symbols, df, dataframe_reddithit_columns):
    from pagexdata.database import db_session
    try:
        from pagexdata.entities import RedditHits
        for subreddit in subreddits:
            for symbol in ticker_symbols:
                key = (df[dataframe_reddithit_columns.COLUMN_IMPORT_DATE] == import_date) & \
                      (df[dataframe_reddithit_columns.COLUMN_TICKER_SYMBOL] == symbol) & \
                      (df[dataframe_reddithit_columns.COLUMN_SUBREDDIT] == subreddit)
                db_session.execute(update(RedditHits)
                    .where(RedditHits.import_date == import_date)
                    .where(RedditHits.subreddit == subreddit)
                    .where(RedditHits.ticker_symbol == symbol)
                    .values(rank=int(df.loc[key][dataframe_reddithit_columns.COLUMN_RANK]),
                            previous_rank=int(df.loc[key][dataframe_reddithit_columns.COLUMN_PREVIOUS_RANK]),
                            change_rank=int(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_RANK]),
                            percentage_ticker_to_total=float(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_RANK]),
                            change_hits_one_day=int(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_HITS_1_DAY]),
                            change_hits_two_days=int(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_HITS_2_DAYS]),
                            change_hits_three_days=int(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_HITS_3_DAYS]),
                            change_hits_one_week=int(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_HITS_1_WEEK]),
                            change_hits_two_weeks=int(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_HITS_2_WEEKS]),
                            change_hits_four_weeks=int(df.loc[key][dataframe_reddithit_columns.COLUMN_CHANGE_HITS_4_WEEKS]),
                            hits_volatility_one_week=float(df.loc[key][dataframe_reddithit_columns.COLUMN_HITS_VOLATILITY_1_WEEK]),
                            hits_volatility_two_weeks=float(df.loc[key][dataframe_reddithit_columns.COLUMN_HITS_VOLATILITY_2_WEEKS])))
        db_session.commit()
        db_session.close()
    except Exception as ex:
        db_session.rollback()
        raise ex



import logging

# setup logger
logging.basicConfig(format="{asctime} {levelname}: pagex-stocks-data - {name}s - {message}",
                    datefmt="%d.%m.%Y %H:%M:%S",
                    style="{", level=logging.INFO,
                    force=True)
logging.getLogger("sqlalchemy").setLevel(logging.ERROR)



import datetime

date_format = "%Y-%m-%d"
datetime_format = "%Y-%m-%d %H:%M:%S"


def create_datetime(custom_date):
    date_to_search = custom_date if custom_date else datetime.date.today() - datetime.timedelta(days=1)
    return datetime.datetime.strptime(str(date_to_search), date_format).replace(tzinfo=datetime.timezone.utc)


def create_date(custom_date):
    return create_datetime(custom_date).date()


def get_str_for_date(custom_date):
    return custom_date.strftime(date_format)


def get_str_for_datetime(custom_datetime):
    return custom_datetime.strftime(datetime_format)
